import sys
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Iterable

from nucleus.annotation import AnnotationList
from nucleus.prediction import PredictionList


@dataclass
class MetricResult:
    """A Metric Result contains the value of an evaluation, as well as its weight.
    The weight is useful when aggregating metrics where each dataset item may hold a
    different relative weight. For example, when calculating precision over a dataset,
    the denominator of the precision is the number of annotations, and therefore the weight
    can be set as the number of annotations.

    Attributes:
        value (float): The value of the evaluation result
        weight (float): The weight of the evaluation result.
    """

    value: float
    weight: float = 1.0

    @staticmethod
    def aggregate(results: Iterable["MetricResult"]) -> "MetricResult":
        """Aggregates results using a weighted average."""
        results = list(filter(lambda x: x.weight != 0, results))
        total_weight = sum([result.weight for result in results])
        total_value = sum([result.value * result.weight for result in results])
        value = total_value / max(total_weight, sys.float_info.epsilon)
        return MetricResult(value, total_weight)


class Metric(ABC):
    """Abstract class for defining a metric, which takes a list of annotations
    and predictions and returns a scalar.

    To create a new concrete Metric, override the `__call__` function
    with logic to define a metric between annotations and predictions. ::

        from nucleus import BoxAnnotation, CuboidPrediction, Point3D
        from nucleus.annotation import AnnotationList
        from nucleus.prediction import PredictionList
        from nucleus.metrics import Metric, MetricResult
        from nucleus.metrics.polygon_utils import BoxOrPolygonAnnotation, BoxOrPolygonPrediction

        class MyMetric(Metric):
            def __call__(
                self, annotations: AnnotationList, predictions: PredictionList
            ) -> MetricResult:
                value = (len(annotations) - len(predictions)) ** 2
                weight = len(annotations)
                return MetricResult(value, weight)

        box = BoxAnnotation(
            label="car",
            x=0,
            y=0,
            width=10,
            height=10,
            reference_id="image_1",
            annotation_id="image_1_car_box_1",
            metadata={"vehicle_color": "red"}
        )

        cuboid = CuboidPrediction(
            label="car",
            position=Point3D(100, 100, 10),
            dimensions=Point3D(5, 10, 5),
            yaw=0,
            reference_id="pointcloud_1",
            confidence=0.8,
            annotation_id="pointcloud_1_car_cuboid_1",
            metadata={"vehicle_color": "green"}
        )

        metric = MyMetric()
        annotations = AnnotationList(box_annotations=[box])
        predictions = PredictionList(cuboid_predictions=[cuboid])
        metric(annotations, predictions)
    """

    @abstractmethod
    def __call__(
        self, annotations: AnnotationList, predictions: PredictionList
    ) -> MetricResult:
        """A metric must override this method and return a metric result, given annotations and predictions."""



"""Unit Tests combine collections of data and evaluation metrics to accelerate model evaluation.

With Model CI Unit Tests, an ML engineer can define a Unit Test from critical
edge case scenarios that the model must get right (e.g. pedestrians at night),
and have confidence that they’re always shipping the best model.
"""
from dataclasses import dataclass, field
from typing import List

from ..connection import Connection
from ..constants import NAME_KEY, SLICE_ID_KEY
from ..dataset_item import DatasetItem
from .data_transfer_objects.eval_function import EvaluationCriterion
from .data_transfer_objects.unit_test_evaluations import GetEvalHistory
from .data_transfer_objects.unit_test_metric import AddUnitTestMetric
from .unit_test_evaluation import UnitTestEvaluation
from .unit_test_metric import UnitTestMetric

DATASET_ITEMS_KEY = "dataset_items"


@dataclass
class UnitTest:
    """A Unit Test combines a slice and at least one evaluation criterion. A :class:`UnitTest` is not created through
    the default constructor but using the instructions shown in :class:`ModelCI`. This :class:`UnitTest` class only
    simplifies the interaction with the unit tests from this SDK.

    Attributes:
        id (str): The ID of the unit test.
        connection (Connection): The connection to Nucleus API.
        name (str): The name of the unit test.
        slice_id (str): The ID of the associated Nucleus slice.
    """

    id: str
    connection: Connection = field(repr=False)
    name: str = field(init=False)
    slice_id: str = field(init=False)

    def __post_init__(self):
        # TODO(gunnar): Remove this pattern. It's too slow. We should get all the info required in one call
        response = self.connection.get(
            f"modelci/unit_test/{self.id}/info",
        )
        self.name = response[NAME_KEY]
        self.slice_id = response[SLICE_ID_KEY]

    def add_criterion(
        self, evaluation_criterion: EvaluationCriterion
    ) -> UnitTestMetric:
        """Creates and adds a new criteria to the :class:`UnitTest`. ::

            import nucleus
            client = nucleus.NucleusClient("YOUR_SCALE_API_KEY")
            unit_test = client.modelci.create_unit_test(
                "sample_unit_test", "slc_bx86ea222a6g057x4380"
            )

            e = client.modelci.eval_functions
            # Assuming a user would like to add all available public evaluation functions as criteria
            unit_test.add_criterion(
                e.bbox_iou() > 0.5
            )
            unit_test.add_criterion(
                e.bbox_map() > 0.85
            )
            unit_test.add_criterion(
                e.bbox_precision() > 0.7
            )
            unit_test.add_criterion(
                e.bbox_recall() > 0.6
            )

        Args:
            evaluation_criterion: :class:`EvaluationCriterion` created by comparison with an :class:`EvalFunction`

        Returns:
            The created UnitTestMetric object.
        """
        response = self.connection.post(
            AddUnitTestMetric(
                unit_test_name=self.name,
                eval_function_id=evaluation_criterion.eval_function_id,
                threshold=evaluation_criterion.threshold,
                threshold_comparison=evaluation_criterion.threshold_comparison,
            ).dict(),
            "modelci/unit_test_metric",
        )
        return UnitTestMetric(
            unit_test_id=response["unit_test_id"],
            eval_function_id=response["eval_function_id"],
            threshold=evaluation_criterion.threshold,
            threshold_comparison=evaluation_criterion.threshold_comparison,
        )

    def get_criteria(self) -> List[UnitTestMetric]:
        """Retrieves all criteria of the :class:`UnitTest`. ::

            import nucleus
            client = nucleus.NucleusClient("YOUR_SCALE_API_KEY")
            unit_test = client.modelci.list_unit_tests()[0]

            unit_test.get_criteria()

        Returns:
            A list of UnitTestMetric objects.
        """
        response = self.connection.get(
            f"modelci/unit_test/{self.id}/metrics",
        )
        return [
            UnitTestMetric(**metric)
            for metric in response["unit_test_metrics"]
        ]

    def get_eval_history(self) -> List[UnitTestEvaluation]:
        """Retrieves evaluation history for :class:`UnitTest`. ::

            import nucleus
            client = nucleus.NucleusClient("YOUR_SCALE_API_KEY")
            unit_test = client.modelci.list_unit_tests()[0]

            unit_test.get_eval_history()

        Returns:
            A list of :class:`UnitTestEvaluation` objects.
        """
        response = self.connection.get(
            f"modelci/unit_test/{self.id}/eval_history",
        )
        eval_history = GetEvalHistory.parse_obj(response)
        return [
            UnitTestEvaluation(evaluation.id, self.connection)
            for evaluation in eval_history.evaluations
        ]

    def get_items(self) -> List[DatasetItem]:
        response = self.connection.get(
            f"modelci/unit_test/{self.id}/items",
        )
        return [
            DatasetItem.from_json(item) for item in response[DATASET_ITEMS_KEY]
        ]



"""Shared stateless utility function library"""

import io
import json
import uuid
from collections import defaultdict
from typing import IO, Dict, List, Sequence, Type, Union

import requests
from requests.models import HTTPError

from nucleus.annotation import (
    Annotation,
    BoxAnnotation,
    CategoryAnnotation,
    CuboidAnnotation,
    MultiCategoryAnnotation,
    PolygonAnnotation,
    SegmentationAnnotation,
)

from .constants import (
    ANNOTATION_TYPES,
    ANNOTATIONS_KEY,
    BOX_TYPE,
    CATEGORY_TYPE,
    CUBOID_TYPE,
    ITEM_KEY,
    MULTICATEGORY_TYPE,
    POLYGON_TYPE,
    REFERENCE_ID_KEY,
    SEGMENTATION_TYPE,
)
from .dataset_item import DatasetItem
from .prediction import (
    BoxPrediction,
    CategoryPrediction,
    CuboidPrediction,
    PolygonPrediction,
    SegmentationPrediction,
)
from .scene import LidarScene

STRING_REPLACEMENTS = {
    "\\\\n": "\n",
    "\\\\t": "\t",
    '\\\\"': '"',
}


class KeyErrorDict(dict):
    """Wrapper for response dicts with deprecated keys.

    Parameters:
        **kwargs: Mapping from the deprecated key to a warning message.
    """

    def __init__(self, **kwargs):
        self._deprecated = {}

        for key, msg in kwargs.items():
            if not isinstance(key, str):
                raise TypeError(
                    f"All keys must be strings! Received non-string '{key}'"
                )
            if not isinstance(msg, str):
                raise TypeError(
                    f"All warning messages must be strings! Received non-string '{msg}'"
                )

            self._deprecated[key] = msg

        super().__init__()

    def __missing__(self, key):
        """Raises KeyError for deprecated keys, otherwise uses base dict logic."""
        if key in self._deprecated:
            raise KeyError(self._deprecated[key])
        try:
            super().__missing__(key)
        except AttributeError as e:
            raise KeyError(key) from e


def format_prediction_response(
    response: dict,
) -> Union[
    dict,
    List[
        Union[
            BoxPrediction,
            PolygonPrediction,
            CuboidPrediction,
            CategoryPrediction,
            SegmentationPrediction,
        ]
    ],
]:
    """Helper function to convert JSON response from endpoints to python objects

    Args:
        response: JSON dictionary response from REST endpoint.
    Returns:
        annotation_response: Dictionary containing a list of annotations for each type,
            keyed by the type name.
    """
    annotation_payload = response.get(ANNOTATIONS_KEY, None)
    if not annotation_payload:
        # An error occurred
        return response
    annotation_response = {}
    type_key_to_class: Dict[
        str,
        Union[
            Type[BoxPrediction],
            Type[PolygonPrediction],
            Type[CuboidPrediction],
            Type[CategoryPrediction],
            Type[SegmentationPrediction],
        ],
    ] = {
        BOX_TYPE: BoxPrediction,
        POLYGON_TYPE: PolygonPrediction,
        CUBOID_TYPE: CuboidPrediction,
        CATEGORY_TYPE: CategoryPrediction,
        SEGMENTATION_TYPE: SegmentationPrediction,
    }
    for type_key in annotation_payload:
        type_class = type_key_to_class[type_key]
        annotation_response[type_key] = [
            type_class.from_json(annotation)
            for annotation in annotation_payload[type_key]
        ]
    return annotation_response


def format_dataset_item_response(response: dict) -> dict:
    """Format the raw client response into api objects.

    Args:
      response: JSON dictionary response from REST endpoint
    Returns:
      item_dict: A dictionary with two entries, one for the dataset item, and annother
        for all of the associated annotations.
    """
    if ANNOTATIONS_KEY not in response:
        raise ValueError(
            f"Server response was missing the annotation key: {response}"
        )
    if ITEM_KEY not in response:
        raise ValueError(
            f"Server response was missing the item key: {response}"
        )
    item = response[ITEM_KEY]
    annotation_payload = response[ANNOTATIONS_KEY]

    annotation_response = {}
    for annotation_type in ANNOTATION_TYPES:
        if annotation_type in annotation_payload:
            annotation_response[annotation_type] = [
                Annotation.from_json(ann)
                for ann in annotation_payload[annotation_type]
            ]
    return {
        ITEM_KEY: DatasetItem.from_json(item),
        ANNOTATIONS_KEY: annotation_response,
    }


def convert_export_payload(api_payload):
    """Helper function to convert raw JSON to API objects

    Args:
        api_payload: JSON dictionary response from REST endpoint
    Returns:
        return_payload: A list of dictionaries for each dataset item. Each dictionary
            is in the same format as format_dataset_item_response: one key for the
            dataset item, another for the annotations.
    """
    return_payload = []
    for row in api_payload:
        return_payload_row = {}
        return_payload_row[ITEM_KEY] = DatasetItem.from_json(row[ITEM_KEY])
        annotations = defaultdict(list)
        if row.get(SEGMENTATION_TYPE) is not None:
            segmentation = row[SEGMENTATION_TYPE]
            segmentation[REFERENCE_ID_KEY] = row[ITEM_KEY][REFERENCE_ID_KEY]
            annotations[SEGMENTATION_TYPE] = SegmentationAnnotation.from_json(
                segmentation
            )
        for polygon in row[POLYGON_TYPE]:
            polygon[REFERENCE_ID_KEY] = row[ITEM_KEY][REFERENCE_ID_KEY]
            annotations[POLYGON_TYPE].append(
                PolygonAnnotation.from_json(polygon)
            )
        for box in row[BOX_TYPE]:
            box[REFERENCE_ID_KEY] = row[ITEM_KEY][REFERENCE_ID_KEY]
            annotations[BOX_TYPE].append(BoxAnnotation.from_json(box))
        for cuboid in row[CUBOID_TYPE]:
            cuboid[REFERENCE_ID_KEY] = row[ITEM_KEY][REFERENCE_ID_KEY]
            annotations[CUBOID_TYPE].append(CuboidAnnotation.from_json(cuboid))
        for category in row[CATEGORY_TYPE]:
            category[REFERENCE_ID_KEY] = row[ITEM_KEY][REFERENCE_ID_KEY]
            annotations[CATEGORY_TYPE].append(
                CategoryAnnotation.from_json(category)
            )
        for multicategory in row[MULTICATEGORY_TYPE]:
            multicategory[REFERENCE_ID_KEY] = row[ITEM_KEY][REFERENCE_ID_KEY]
            annotations[MULTICATEGORY_TYPE].append(
                MultiCategoryAnnotation.from_json(multicategory)
            )
        return_payload_row[ANNOTATIONS_KEY] = annotations
        return_payload.append(return_payload_row)
    return return_payload


def serialize_and_write(
    upload_units: Sequence[Union[DatasetItem, Annotation, LidarScene]],
    file_pointer,
):
    if len(upload_units) == 0:
        raise ValueError(
            "Expecting at least one object when serializing objects to upload, but got zero.  Please try again."
        )
    for unit in upload_units:
        try:
            if isinstance(unit, (DatasetItem, Annotation, LidarScene)):
                file_pointer.write(unit.to_json() + "\n")
            else:
                file_pointer.write(json.dumps(unit) + "\n")
        except TypeError as e:
            type_name = type(unit).__name__
            message = (
                f"The following {type_name} could not be serialized: {unit}\n"
            )
            message += (
                "This is usally an issue with a custom python object being "
                "present in the metadata. Please inspect this error and adjust the "
                "metadata so it is json-serializable: only python primitives such as "
                "strings, ints, floats, lists, and dicts. For example, you must "
                "convert numpy arrays into list or lists of lists.\n"
            )
            message += f"The specific error was {e}"
            raise ValueError(message) from e


def upload_to_presigned_url(presigned_url: str, file_pointer: IO):
    # TODO optimize this further to deal with truly huge files and flaky internet connection.
    upload_response = requests.put(presigned_url, file_pointer)
    if not upload_response.ok:
        raise HTTPError(
            f"Tried to put a file to url, but failed with status {upload_response.status_code}. The detailed error was: {upload_response.text}"
        )


def serialize_and_write_to_presigned_url(
    upload_units: Sequence[Union[DatasetItem, Annotation, LidarScene]],
    dataset_id: str,
    client,
):
    """This helper function can be used to serialize a list of API objects to NDJSON."""
    request_id = uuid.uuid4().hex
    response = client.make_request(
        payload={},
        route=f"dataset/{dataset_id}/signedUrl/{request_id}",
        requests_command=requests.get,
    )

    strio = io.StringIO()
    serialize_and_write(upload_units, strio)
    strio.seek(0)
    upload_to_presigned_url(response["signed_url"], strio)
    return request_id


def replace_double_slashes(s: str) -> str:
    for key, val in STRING_REPLACEMENTS.items():
        s = s.replace(key, val)
    return s



from typing import List, Tuple, Union

import numpy as np

TOLERANCE = 1e-8


class GeometryPoint:
    def __init__(self, xy: Union[Tuple[float, float], np.ndarray]):
        self.xy = np.array(xy)
        self.x = xy[0]
        self.y = xy[1]

    def __repr__(self) -> str:
        return f"GeometryPoint(xy=({self.xy[0]}, {self.xy[1]})"

    def __add__(self, p: "GeometryPoint") -> "GeometryPoint":
        return GeometryPoint(self.xy + p.xy)

    def __sub__(self, p: "GeometryPoint") -> "GeometryPoint":
        return GeometryPoint(self.xy - p.xy)

    def __rmul__(self, scalar: float) -> "GeometryPoint":
        return GeometryPoint(self.xy * scalar)

    def __mul__(self, scalar: float) -> "GeometryPoint":
        return GeometryPoint(self.xy * scalar)

    # Operator @
    def __matmul__(self, p: "GeometryPoint") -> float:
        return self.xy @ p.xy

    def length(self) -> float:
        return np.linalg.norm(self.xy)

    def cmp(self, p: "GeometryPoint") -> float:
        return np.abs(self.xy - p.xy).sum()


class GeometryPolygon:
    def __init__(self, points: List[GeometryPoint]):
        self.points = points

        points_x = np.array([point.x for point in points])
        points_y = np.array([point.y for point in points])
        self.signed_area = (
            points_x @ np.roll(points_y, 1) - points_x @ np.roll(points_y, -1)
        ) / 2
        self.area = np.abs(self.signed_area)

    def __len__(self):
        return len(self.points)

    def __getitem__(self, idx):
        return self.points[idx]

    def __repr__(self) -> str:
        points = ", ".join([str(point) for point in self.points])
        return f"GeometryPolygon({points})"


Segment = Tuple[GeometryPoint, GeometryPoint]


# alpha * a1 + (1 - alpha) * a2 = beta * b1 + (1 - beta) * b2
def segment_intersection(
    segment1: Segment, segment2: Segment
) -> Tuple[float, float, GeometryPoint]:
    a1, a2 = segment1
    b1, b2 = segment2
    x2_x2 = b2.x - a2.x
    y2_y2 = b2.y - a2.y
    x1x2 = a1.x - a2.x
    y1y2 = a1.y - a2.y
    y1_y2_ = b1.y - b2.y
    x1_x2_ = b1.x - b2.x

    if np.abs(y1_y2_ * x1x2 - x1_x2_ * y1y2) < TOLERANCE:
        beta = 1.0
    else:
        beta = (x2_x2 * y1y2 - y2_y2 * x1x2) / (y1_y2_ * x1x2 - x1_x2_ * y1y2)

    if x1x2 == 0:
        alpha = (y2_y2 + y1_y2_ * beta) / (y1y2 + TOLERANCE)
    else:
        alpha = (x2_x2 + x1_x2_ * beta) / (x1x2 + TOLERANCE)

    return alpha, beta, alpha * a1 + (1 - alpha) * a2


def convex_polygon_intersection_area(
    polygon_a: GeometryPolygon, polygon_b: GeometryPolygon
) -> float:
    # pylint: disable=R0912
    sa = polygon_a.signed_area
    sb = polygon_b.signed_area
    if sa * sb < 0:
        sign = -1
    else:
        sign = 1
    na = len(polygon_a)
    nb = len(polygon_b)
    ps = []  # point set
    for i in range(na):
        a1 = polygon_a[i - 1]
        a2 = polygon_a[i]
        flag = False
        sum_s = 0
        for j in range(nb):
            b1 = polygon_b[j - 1]
            b2 = polygon_b[j]
            sum_s += np.abs(GeometryPolygon([a1, b1, b2]).signed_area)

        if np.abs(np.abs(sum_s) - np.abs(sb)) < TOLERANCE:
            flag = True

        if flag:
            ps.append(a1)
        for j in range(nb):
            b1 = polygon_b[j - 1]
            b2 = polygon_b[j]
            a, b, p = segment_intersection((a1, a2), (b1, b2))
            if 0 < a < 1 and 0 < b < 1:
                ps.append(p)

    for i in range(nb):
        a1 = polygon_b[i - 1]
        a2 = polygon_b[i]
        flag = False
        sum_s = 0
        for j in range(na):
            b1 = polygon_a[j - 1]
            b2 = polygon_a[j]
            sum_s += np.abs(GeometryPolygon([a1, b1, b2]).signed_area)
        if np.abs(np.abs(sum_s) - np.abs(sa)) < TOLERANCE:
            flag = True
        if flag:
            ps.append(a1)

    def unique(ar):
        res = []
        for i, _ in enumerate(ar):
            if _.cmp(ar[i - 1]) > TOLERANCE:
                res.append(_)

        return res

    ps = sorted(ps, key=lambda x: (x.x + TOLERANCE * x.y))
    ps = unique(ps)

    if len(ps) == 0:
        return 0

    tmp = ps[0]

    res = []
    res.append(tmp)
    ps = sorted(
        ps[1:],
        key=lambda x: -(
            (x - tmp) @ GeometryPoint((0, 1)) / (x - tmp).length()
        ),
    )
    res.extend(ps)

    return GeometryPolygon(res).signed_area * sign


def polygon_intersection_area(
    polygon_a: GeometryPolygon, polygon_b: GeometryPolygon
) -> float:
    na = len(polygon_a)
    nb = len(polygon_b)
    res = 0.0
    for i in range(1, na - 1):
        sa = [polygon_a[0], polygon_a[i], polygon_a[i + 1]]
        for j in range(1, nb - 1):
            sb = [polygon_b[0], polygon_b[j], polygon_b[j + 1]]
            tmp = convex_polygon_intersection_area(
                GeometryPolygon(sa), GeometryPolygon(sb)
            )
            res += tmp

    return np.abs(res)



import itertools
from typing import Callable, Dict, List, Type, Union

from nucleus.logger import logger
from nucleus.modelci.eval_functions.base_eval_function import BaseEvalFunction

from ..data_transfer_objects.eval_function import EvalFunctionEntry
from ..errors import EvalFunctionNotAvailableError

MEAN_AVG_PRECISION_NAME = "mean_average_precision_boxes"


class BoundingBoxIOU(BaseEvalFunction):
    @classmethod
    def expected_name(cls) -> str:
        return "bbox_iou"


class BoundingBoxMeanAveragePrecision(BaseEvalFunction):
    @classmethod
    def expected_name(cls) -> str:
        return "bbox_map"


class BoundingBoxRecall(BaseEvalFunction):
    @classmethod
    def expected_name(cls) -> str:
        return "bbox_recall"


class BoundingBoxPrecision(BaseEvalFunction):
    @classmethod
    def expected_name(cls) -> str:
        return "bbox_precision"


class CustomEvalFunction(BaseEvalFunction):
    @classmethod
    def expected_name(cls) -> str:
        raise NotImplementedError(
            "Custm evaluation functions are coming soon"
        )  # Placeholder: See super().eval_func_entry for actual name


class StandardEvalFunction(BaseEvalFunction):
    """Class for standard Model CI eval functions that have not been added as attributes on
    AvailableEvalFunctions yet.
    """

    def __init__(self, eval_function_entry: EvalFunctionEntry):
        logger.warning(
            "Standard function %s not implemented as an attribute on AvailableEvalFunctions",
            eval_function_entry.name,
        )
        super().__init__(eval_function_entry)

    @classmethod
    def expected_name(cls) -> str:
        return "public_function"  # Placeholder: See super().eval_func_entry for actual name


class EvalFunctionNotAvailable(BaseEvalFunction):
    def __init__(
        self, not_available_name: str
    ):  # pylint: disable=super-init-not-called
        self.not_available_name = not_available_name

    def __call__(self, *args, **kwargs):
        self._raise_error()

    def _op_to_test_metric(self, *args, **kwargs):
        self._raise_error()

    def _raise_error(self):
        raise EvalFunctionNotAvailableError(
            f"Eval function '{self.not_available_name}' is not available to the current user. "
            f"Is Model CI enabled for the user?"
        )

    @classmethod
    def expected_name(cls) -> str:
        return "public_function"  # Placeholder: See super().eval_func_entry for actual name


EvalFunction = Union[
    Type[BoundingBoxIOU],
    Type[BoundingBoxMeanAveragePrecision],
    Type[BoundingBoxPrecision],
    Type[BoundingBoxRecall],
    Type[CustomEvalFunction],
    Type[EvalFunctionNotAvailable],
    Type[StandardEvalFunction],
]


class AvailableEvalFunctions:
    """Collection class that acts as a common entrypoint to access evaluation functions. Standard evaluation functions
    provided by Scale are attributes of this class.

    The available evaluation functions are listed in the sample below::

        e = client.modelci.eval_functions
        unit_test_criteria = [
            e.bbox_iou() > 5,
            e.bbox_map() > 0.95,
            e.bbox_precision() > 0.8,
            e.bbox_recall() > 0.5,
        ]
    """

    # pylint: disable=too-many-instance-attributes

    def __init__(self, available_functions: List[EvalFunctionEntry]):
        assert (
            available_functions
        ), "Passed no available functions for current user. Is the feature flag enabled?"
        self._public_func_entries: Dict[str, EvalFunctionEntry] = {
            f.name: f for f in available_functions if f.is_public
        }
        # NOTE: Public are assigned
        self._public_to_function: Dict[str, BaseEvalFunction] = {}
        self._custom_to_function: Dict[str, CustomEvalFunction] = {
            f.name: CustomEvalFunction(f)
            for f in available_functions
            if not f.is_public
        }
        self.bbox_iou = self._assign_eval_function_if_defined(BoundingBoxIOU)  # type: ignore
        self.bbox_precision = self._assign_eval_function_if_defined(
            BoundingBoxPrecision  # type: ignore
        )
        self.bbox_recall = self._assign_eval_function_if_defined(
            BoundingBoxRecall  # type: ignore
        )
        self.bbox_map = self._assign_eval_function_if_defined(
            BoundingBoxMeanAveragePrecision  # type: ignore
        )

        # Add public entries that have not been implemented as an attribute on this class
        for func_entry in self._public_func_entries.values():
            if func_entry.name not in self._public_to_function:
                self._public_to_function[
                    func_entry.name
                ] = StandardEvalFunction(func_entry)

    def __repr__(self):
        """Standard functions are ones Scale provides and custom ones customer defined"""
        # NOTE: setting to lower to be consistent with attribute names
        functions_lower = [
            str(name).lower() for name in self._public_func_entries.keys()
        ]
        return (
            f"<AvailableEvaluationFunctions: public:{functions_lower} "
            f"private: {list(self._custom_to_function.keys())}"
        )

    @property
    def public_functions(self) -> Dict[str, BaseEvalFunction]:
        """Standard functions provided by Model CI.

        Notes:
            These functions are also available as attributes on :class:`AvailableEvalFunctions`

        Returns:
            Dict of function name to :class:`BaseEvalFunction`.
        """
        return self._public_to_function

    @property
    def private_functions(self) -> Dict[str, CustomEvalFunction]:
        """Custom functions uploaded to Model CI

        Returns:
            Dict of function name to :class:`CustomEvalFunction`.
        """
        return self._custom_to_function

    def _assign_eval_function_if_defined(
        self,
        eval_function_constructor: Callable[[EvalFunctionEntry], EvalFunction],
    ):
        """Helper function for book-keeping and assignment of standard Scale provided functions that are accessible
        via attribute access
        """
        # TODO(gunnar): Too convoluted .. simplify
        expected_name = eval_function_constructor.expected_name()  # type: ignore
        if expected_name in self._public_func_entries:
            definition = self._public_func_entries[expected_name]
            eval_function = eval_function_constructor(definition)
            self._public_to_function[expected_name] = eval_function  # type: ignore
            return eval_function
        else:
            return EvalFunctionNotAvailable(expected_name)

    def from_id(self, eval_function_id: str):
        for eval_func in itertools.chain(
            self._public_to_function.values(),
            self._custom_to_function.values(),
        ):
            if eval_func.id == eval_function_id:
                return eval_func
        raise EvalFunctionNotAvailableError(
            f"Could not find Eval Function with id {eval_function_id}"
        )



from nucleus.pydantic_base import ImmutableModel

from ..constants import ThresholdComparison


class AddUnitTestMetric(ImmutableModel):
    """Data transfer object to add a unit test."""

    unit_test_name: str
    eval_function_id: str
    threshold: float
    threshold_comparison: ThresholdComparison



from nucleus.pydantic_base import ImmutableModel

from .constants import ThresholdComparison


class UnitTestMetric(ImmutableModel):
    """A Unit Test Metric is an evaluation function combined with a comparator and associated with a Unit Test.
    Unit Test Metrics serve as the basis when evaluating a Model on a Unit Test.
    """

    unit_test_id: str
    eval_function_id: str
    threshold: float
    threshold_comparison: ThresholdComparison



"""Model CI Python Library."""

__all__ = [
    "ModelCI",
    "UnitTest",
    "EvaluationCriterion",
]

from .client import ModelCI
from .constants import ThresholdComparison
from .data_transfer_objects.eval_function import (
    EvalFunctionEntry,
    EvaluationCriterion,
    GetEvalFunctions,
)
from .data_transfer_objects.unit_test import CreateUnitTestRequest
from .errors import CreateUnitTestError
from .eval_functions.available_eval_functions import AvailableEvalFunctions
from .unit_test import UnitTest
from .unit_test_evaluation import UnitTestEvaluation, UnitTestItemEvaluation
from .unit_test_metric import UnitTestMetric



from typing import List

from nucleus.connection import Connection
from nucleus.job import AsyncJob

from .constants import UNIT_TEST_ID_KEY
from .data_transfer_objects.eval_function import (
    EvaluationCriterion,
    GetEvalFunctions,
)
from .data_transfer_objects.unit_test import CreateUnitTestRequest
from .errors import CreateUnitTestError
from .eval_functions.available_eval_functions import AvailableEvalFunctions
from .unit_test import UnitTest

SUCCESS_KEY = "success"
EVAL_FUNCTIONS_KEY = "eval_functions"


class ModelCI:
    """Model CI Python Client extension."""

    def __init__(self, api_key: str, endpoint: str):
        self.connection = Connection(api_key, endpoint)

    def __repr__(self):
        return f"ModelCI(connection='{self.connection}')"

    def __eq__(self, other):
        return self.connection == other.connection

    @property
    def eval_functions(self) -> AvailableEvalFunctions:
        """List all available evaluation functions which can be used to set up evaluation criteria.::

            import nucleus
            client = nucleus.NucleusClient("YOUR_SCALE_API_KEY")

            unit_test_criterion = client.modelci.eval_functions.bbox_iou() > 0.5  # Creates an EvaluationCriterion by comparison

        Returns:
            :class:`AvailableEvalFunctions`: A container for all the available eval functions
        """
        response = self.connection.get(
            "modelci/eval_fn",
        )
        payload = GetEvalFunctions.parse_obj(response)
        return AvailableEvalFunctions(payload.eval_functions)

    def create_unit_test(
        self,
        name: str,
        slice_id: str,
        evaluation_criteria: List[EvaluationCriterion],
    ) -> UnitTest:
        """Creates a new Unit Test from an existing Nucleus :class:`Slice`:. ::

            import nucleus
            client = nucleus.NucleusClient("YOUR_SCALE_API_KEY")

            unit_test = client.modelci.create_unit_test(
                name="sample_unit_test",
                slice_id="YOUR_SLICE_ID",
                evaluation_criteria=[client.modelci.eval_functions.bbox_iou() > 0.5]
            )

        Args:
            name: unique name of test
            slice_id: id of (pre-defined) slice of items to evaluate test on.
            evaluation_criteria: :class:`EvaluationCriterion` defines a pass/fail criteria for the test. Created with a
                comparison with an eval functions. See :class:`eval_functions`.

        Returns:
            Created UnitTest object.
        """
        if not evaluation_criteria:
            raise CreateUnitTestError(
                "Must pass an evaluation_criteria to the unit test! I.e. "
                "evaluation_criteria = [client.modelci.eval_functions.bbox_iou() > 0.5]"
            )
        response = self.connection.post(
            CreateUnitTestRequest(
                name=name,
                slice_id=slice_id,
                evaluation_criteria=evaluation_criteria,
            ).dict(),
            "modelci/unit_test",
        )
        return UnitTest(response[UNIT_TEST_ID_KEY], self.connection)

    def get_unit_test(self, unit_test_id: str) -> UnitTest:
        response = self.connection.get(
            f"modelci/unit_test/{unit_test_id}",
        )
        return UnitTest(response["id"], self.connection)

    def list_unit_tests(self) -> List[UnitTest]:
        """Lists all Unit Tests of the current user. ::

            import nucleus
            client = nucleus.NucleusClient("YOUR_SCALE_API_KEY")
            unit_test = client.modelci.create_unit_test(
                "sample_unit_test", "slc_bx86ea222a6g057x4380"
            )

            client.modelci.list_unit_tests()

        Returns:
            A list of UnitTest objects.
        """
        response = self.connection.get(
            "modelci/unit_test",
        )
        return [
            UnitTest(test_id, self.connection)
            for test_id in response["unit_test_ids"]
        ]

    def delete_unit_test(self, unit_test_id: str) -> bool:
        """Deletes a Unit Test. ::

            import nucleus
            client = nucleus.NucleusClient("YOUR_SCALE_API_KEY")
            unit_test = client.modelci.list_unit_tests()[0]

            success = client.modelci.delete_unit_test(unit_test.id)

        Args:
            unit_test_id: unique ID of unit test

        Returns:
            Whether deletion was successful.
        """
        response = self.connection.delete(
            f"modelci/unit_test/{unit_test_id}",
        )
        return response[SUCCESS_KEY]

    def evaluate_model_on_unit_tests(
        self, model_id: str, unit_test_names: List[str]
    ) -> AsyncJob:
        """Evaluates the given model on the specified Unit Tests. ::

            import nucleus
            client = nucleus.NucleusClient("YOUR_SCALE_API_KEY")
            model = client.list_models()[0]
            unit_test = client.modelci.create_unit_test(
                "sample_unit_test", "slc_bx86ea222a6g057x4380"
            )

            job = client.modelci.evaluate_model_on_unit_tests(
                model_id=model.id,
                unit_test_names=["sample_unit_test"],
            )
            job.sleep_until_complete() # Not required. Will block and update on status of the job.

        Args:
            model_id: ID of model to evaluate
            unit_test_names: list of unit test names of test to evaluate

        Returns:
            AsyncJob object of evaluation job
        """
        response = self.connection.post(
            {"test_names": unit_test_names},
            f"modelci/{model_id}/evaluate",
        )
        return AsyncJob.from_json(response, self.connection)



import asyncio
import json
import os
import time
from typing import TYPE_CHECKING, Any, List

import aiohttp
import nest_asyncio

from .constants import (
    DATASET_ID_KEY,
    DEFAULT_NETWORK_TIMEOUT_SEC,
    IMAGE_KEY,
    IMAGE_URL_KEY,
    ITEMS_KEY,
    UPDATE_KEY,
)
from .dataset_item import DatasetItem
from .errors import NotFoundError
from .logger import logger
from .payload_constructor import construct_append_payload
from .retry_strategy import RetryStrategy
from .upload_response import UploadResponse

if TYPE_CHECKING:
    from . import NucleusClient


class DatasetItemUploader:
    def __init__(self, dataset_id: str, client: "NucleusClient"):  # noqa: F821
        self.dataset_id = dataset_id
        self._client = client

    def upload(
        self,
        dataset_items: List[DatasetItem],
        batch_size: int = 20,
        update: bool = False,
    ) -> UploadResponse:
        """

        Args:
            dataset_items: Items to Upload
            batch_size: How many items to pool together for a single request
            update: Update records instead of overwriting

        Returns:

        """
        local_items = []
        remote_items = []

        # Check local files exist before sending requests
        for item in dataset_items:
            if item.local:
                if not item.local_file_exists():
                    raise NotFoundError()
                local_items.append(item)
            else:
                remote_items.append(item)

        local_batches = [
            local_items[i : i + batch_size]
            for i in range(0, len(local_items), batch_size)
        ]

        remote_batches = [
            remote_items[i : i + batch_size]
            for i in range(0, len(remote_items), batch_size)
        ]

        agg_response = UploadResponse(json={DATASET_ID_KEY: self.dataset_id})

        async_responses: List[Any] = []

        if local_batches:
            tqdm_local_batches = self._client.tqdm_bar(
                local_batches, desc="Local file batches"
            )

            for batch in tqdm_local_batches:
                payload = construct_append_payload(batch, update)
                responses = self._process_append_requests_local(
                    self.dataset_id, payload, update
                )
                async_responses.extend(responses)

        if remote_batches:
            tqdm_remote_batches = self._client.tqdm_bar(
                remote_batches, desc="Remote file batches"
            )
            for batch in tqdm_remote_batches:
                payload = construct_append_payload(batch, update)
                responses = self._process_append_requests(
                    dataset_id=self.dataset_id,
                    payload=payload,
                    update=update,
                    batch_size=batch_size,
                )
                async_responses.extend(responses)

        for response in async_responses:
            agg_response.update_response(response)

        return agg_response

    def _process_append_requests_local(
        self,
        dataset_id: str,
        payload: dict,
        update: bool,  # TODO: understand how to pass this in.
        local_batch_size: int = 10,
    ):
        def get_files(batch):
            for item in batch:
                item[UPDATE_KEY] = update
            request_payload = [
                (
                    ITEMS_KEY,
                    (
                        None,
                        json.dumps(batch, allow_nan=False),
                        "application/json",
                    ),
                )
            ]
            for item in batch:
                image = open(  # pylint: disable=R1732
                    item.get(IMAGE_URL_KEY), "rb"  # pylint: disable=R1732
                )  # pylint: disable=R1732
                img_name = os.path.basename(image.name)
                img_type = (
                    f"image/{os.path.splitext(image.name)[1].strip('.')}"
                )
                request_payload.append(
                    (IMAGE_KEY, (img_name, image, img_type))
                )
            return request_payload

        items = payload[ITEMS_KEY]
        responses: List[Any] = []
        files_per_request = []
        payload_items = []
        for i in range(0, len(items), local_batch_size):
            batch = items[i : i + local_batch_size]
            files_per_request.append(get_files(batch))
            payload_items.append(batch)

        future = self.make_many_files_requests_asynchronously(
            files_per_request,
            f"dataset/{dataset_id}/append",
        )

        try:
            loop = asyncio.get_event_loop()
        except RuntimeError:  # no event loop running:
            loop = asyncio.new_event_loop()
            responses = loop.run_until_complete(future)
        else:
            nest_asyncio.apply(loop)
            return loop.run_until_complete(future)

        def close_files(request_items):
            for item in request_items:
                # file buffer in location [1][1]
                if item[0] == IMAGE_KEY:
                    item[1][1].close()

        # don't forget to close all open files
        for p in files_per_request:
            close_files(p)

        return responses

    async def make_many_files_requests_asynchronously(
        self, files_per_request, route
    ):
        """
        Makes an async post request with files to a Nucleus endpoint.

        :param files_per_request: A list of lists of tuples (name, (filename, file_pointer, content_type))
           name will become the name by which the multer can build an array.
        :param route: route for the request
        :return: awaitable list(response)
        """
        async with aiohttp.ClientSession() as session:
            tasks = [
                asyncio.ensure_future(
                    self._make_files_request(
                        files=files, route=route, session=session
                    )
                )
                for files in files_per_request
            ]
            return await asyncio.gather(*tasks)

    async def _make_files_request(
        self,
        files,
        route: str,
        session: aiohttp.ClientSession,
        retry_attempt=0,
        max_retries=3,
        sleep_intervals=(1, 3, 9),
    ):
        """
        Makes an async post request with files to a Nucleus endpoint.

        :param files: A list of tuples (name, (filename, file_pointer, file_type))
        :param route: route for the request
        :param session: Session to use for post.
        :return: response
        """
        endpoint = f"{self._client.endpoint}/{route}"

        logger.info("Posting to %s", endpoint)

        form = aiohttp.FormData()

        for file in files:
            form.add_field(
                name=file[0],
                filename=file[1][0],
                value=file[1][1],
                content_type=file[1][2],
            )

        for sleep_time in RetryStrategy.sleep_times + [-1]:

            async with session.post(
                endpoint,
                data=form,
                auth=aiohttp.BasicAuth(self._client.api_key, ""),
                timeout=DEFAULT_NETWORK_TIMEOUT_SEC,
            ) as response:
                logger.info(
                    "API request has response code %s", response.status
                )

                try:
                    data = await response.json()
                except aiohttp.client_exceptions.ContentTypeError:
                    # In case of 404, the server returns text
                    data = await response.text()
                if (
                    response.status in RetryStrategy.statuses
                    and sleep_time != -1
                ):
                    time.sleep(sleep_time)
                    continue

                if not response.ok:
                    if retry_attempt < max_retries:
                        time.sleep(sleep_intervals[retry_attempt])
                        retry_attempt += 1
                        return self._make_files_request(
                            files,
                            route,
                            session,
                            retry_attempt,
                            max_retries,
                            sleep_intervals,
                        )
                    else:
                        self._client.handle_bad_response(
                            endpoint,
                            session.post,
                            aiohttp_response=(
                                response.status,
                                response.reason,
                                data,
                            ),
                        )

                return data

    def _process_append_requests(
        self,
        dataset_id: str,
        payload: dict,
        update: bool,
        batch_size: int = 20,
    ):
        items = payload[ITEMS_KEY]
        payloads = [
            # batch_size images per request
            {ITEMS_KEY: items[i : i + batch_size], UPDATE_KEY: update}
            for i in range(0, len(items), batch_size)
        ]

        return [
            self._client.make_request(
                payload,
                f"dataset/{dataset_id}/append",
            )
            for payload in payloads
        ]
